@@ -33,11 +33,13 @@
 #include <xalloc.h>
 #include <gl_linked_list.h>
 #include <gl_xlist.h>
+#include <minmax.h>
 
 /* procedures */
 
 static FILE *create_output_file (char const *, int);
 static lin locate_hunk (lin);
+static bool check_line_endings (lin);
 static bool apply_hunk (struct outstate *, lin);
 static bool patch_match (lin, lin, lin, lin);
 static bool spew_output (struct outstate *, struct stat *);
@@ -48,14 +50,14 @@ static void init_output (struct outstate *);
 static FILE *open_outfile (char const *);
 static void init_reject (char const *);
 static void reinitialize_almost_everything (void);
-static void remove_if_needed (char const *, int *);
+static void remove_if_needed (char const *, bool *);
 static void usage (FILE *, int) __attribute__((noreturn));
 
 static void abort_hunk (char const *, bool, bool);
 static void abort_hunk_context (bool, bool);
 static void abort_hunk_unified (bool, bool);
 
-static void output_file (char const *, int *, const struct stat *, char const *,
+static void output_file (char const *, bool *, const struct stat *, char const *,
 			 const struct stat *, mode_t, bool);
 
 static void init_files_to_delete (void);
@@ -95,7 +97,7 @@ static FILE *rejfp;  /* reject file pointer */
 static char const *patchname;
 static char *rejname;
 static char const * TMPREJNAME;
-static int TMPREJNAME_needs_removal;
+static bool TMPREJNAME_needs_removal;
 
 static lin maxfuzz = 2;
 
@@ -149,6 +151,10 @@ main (int argc, char **argv)
     else if ((version_control = getenv ("VERSION_CONTROL")))
       version_control_context = "$VERSION_CONTROL";
 
+    init_backup_hash_table ();
+    init_files_to_delete ();
+    init_files_to_output ();
+
     /* parse switches */
     Argc = argc;
     Argv = argv;
@@ -161,10 +167,6 @@ main (int argc, char **argv)
     if (make_backups | backup_if_mismatch)
       backup_type = get_version (version_control_context, version_control);
 
-    init_backup_hash_table ();
-    init_files_to_delete ();
-    init_files_to_output ();
-
     init_output (&outstate);
     if (outfile)
       outstate.ofp = open_outfile (outfile);
@@ -196,8 +198,12 @@ main (int argc, char **argv)
 
       if (have_git_diff != pch_git_diff ())
 	{
+	  if (have_git_diff)
+	    {
+	      output_files (NULL);
+	      inerrno = -1;
+	    }
 	  have_git_diff = ! have_git_diff;
-	  output_files (NULL);
 	}
 
       if (TMPREJNAME_needs_removal)
@@ -251,19 +257,19 @@ main (int argc, char **argv)
 	  if (! strcmp (inname, outname))
 	    {
 	      if (inerrno == -1)
-		inerrno = lstat (inname, &instat) ? errno : 0;
+		inerrno = stat_file (inname, &instat);
 	      outstat = instat;
 	      outerrno = inerrno;
 	    }
 	  else
-	    outerrno = lstat (outname, &outstat) ? errno : 0;
+	    outerrno = stat_file (outname, &outstat);
 
 	  if (! outerrno)
 	    {
 	      if (has_queued_output (&outstat))
 		{
 		  output_files (&outstat);
-		  outerrno = lstat (outname, &outstat) ? errno : 0;
+		  outerrno = stat_file (outname, &outstat);
 		  inerrno = -1;
 		}
 	      if (! outerrno)
@@ -297,8 +303,9 @@ main (int argc, char **argv)
 
       tmpoutst.st_size = -1;
       outfd = make_tempfile (&TMPOUTNAME, 'o', outname,
-			     O_WRONLY | binary_transput, instat.st_mode);
-      TMPOUTNAME_needs_removal = 1;
+			     O_WRONLY | binary_transput,
+			     instat.st_mode & S_IRWXUGO);
+      TMPOUTNAME_needs_removal = true;
       if (diff_type == ED_DIFF) {
 	outstate.zero_output = false;
 	somefailed |= skip_rest_of_patch;
@@ -339,7 +346,8 @@ main (int argc, char **argv)
 	      {
 		bool renamed = strcmp (inname, outname);
 
-		say ("patching %s %s%c",
+		say ("%s %s %s%c",
+		     dry_run ? "checking" : "patching",
 		     S_ISLNK (file_type) ? "symbolic link" : "file",
 		     quotearg (outname), renamed ? ' ' : '\n');
 		if (renamed)
@@ -440,9 +448,11 @@ main (int argc, char **argv)
 		failed++;
 		if (verbosity == VERBOSE ||
 		    (! skip_rest_of_patch && verbosity != SILENT))
-		  say ("Hunk #%d %s at %s.\n", hunk,
+		  say ("Hunk #%d %s at %s%s.\n", hunk,
 		       skip_rest_of_patch ? "ignored" : "FAILED",
-		       format_linenum (numbuf, newwhere));
+		       format_linenum (numbuf, newwhere),
+		       ! skip_rest_of_patch && check_line_endings (newwhere)
+			 ?  " (different line endings)" : "");
 	      }
 	    else if (! merge &&
 		     (verbosity == VERBOSE
@@ -503,12 +513,14 @@ main (int argc, char **argv)
 	  else
 	    {
 	      if (! outstate.zero_output
-		  && pch_says_nonexistent (! reverse)
+		  && pch_says_nonexistent (! reverse) == 2
+		  && (remove_empty_files || ! posixly_correct)
 		  && ! (merge && somefailed))
 		{
 		  mismatch = true;
+		  somefailed = true;
 		  if (verbosity != SILENT)
-		    say ("File %s is not empty after patch, as expected\n",
+		    say ("Not deleting file %s as content differs from patch\n",
 			 quotearg (outname));
 		}
 
@@ -593,9 +605,9 @@ main (int argc, char **argv)
 		      rej[len - 1] = '#';
 		    simple_backup_suffix = s;
 		}
-		say (" -- saving rejects to file %s\n", quotearg (rej));
 		if (! dry_run)
 		  {
+		    say (" -- saving rejects to file %s\n", quotearg (rej));
 		    if (rejname)
 		      {
 			if (! written_to_rejname)
@@ -612,7 +624,7 @@ main (int argc, char **argv)
 			struct stat oldst;
 			int olderrno;
 
-			olderrno = lstat (rej, &oldst) ? errno : 0;
+			olderrno = stat_file (rej, &oldst);
 			if (olderrno && olderrno != ENOENT)
 			  write_fatal ();
 		        if (! olderrno && lookup_file_id (&oldst) == CREATED)
@@ -622,6 +634,8 @@ main (int argc, char **argv)
 				     &rejst, rej, S_IFREG | 0666, false);
 		      }
 		  }
+		else
+		  say ("\n");
 		if (!rejname)
 		    free (rej);
 	    } else
@@ -633,8 +647,8 @@ main (int argc, char **argv)
     if (outstate.ofp && (ferror (outstate.ofp) || fclose (outstate.ofp) != 0))
       write_fatal ();
     output_files (NULL);
-    delete_files ();
     cleanup ();
+    delete_files ();
     if (somefailed)
       exit (1);
     return 0;
@@ -720,6 +734,7 @@ static struct option const longopts[] =
   {"quoting-style", required_argument, NULL, CHAR_MAX + 8},
   {"reject-format", required_argument, NULL, CHAR_MAX + 9},
   {"read-only", required_argument, NULL, CHAR_MAX + 10},
+  {"follow-symlinks", no_argument, NULL, CHAR_MAX + 11},
   {NULL, no_argument, NULL, 0}
 };
 
@@ -748,7 +763,7 @@ static char const *const option_help[] =
 "",
 "  -D NAME  --ifdef=NAME  Make merged if-then-else output using NAME.",
 #ifdef ENABLE_MERGE
-"  -m  --merge  Merge using conflict markers instead of creating reject files.",
+"  --merge  Merge using conflict markers instead of creating reject files.",
 #endif
 "  -E  --remove-empty-files  Remove output files that are empty after patching.",
 "",
@@ -851,7 +866,7 @@ get_some_switches (void)
 	    case 'B':
 		if (!*optarg)
 		  fatal ("backup prefix is empty");
-		origprae = savestr (optarg);
+		origprae = xstrdup (optarg);
 		break;
 	    case 'c':
 		diff_type = CONTEXT_DIFF;
@@ -861,7 +876,7 @@ get_some_switches (void)
 		  pfatal ("Can't change to directory %s", quotearg (optarg));
 		break;
 	    case 'D':
-		do_defines = savestr (optarg);
+		do_defines = xstrdup (optarg);
 		break;
 	    case 'e':
 		diff_type = ED_DIFF;
@@ -879,7 +894,7 @@ get_some_switches (void)
 		patch_get = numeric_string (optarg, true, "get option value");
 		break;
 	    case 'i':
-		patchname = savestr (optarg);
+		patchname = xstrdup (optarg);
 		break;
 	    case 'l':
 		canonicalize = true;
@@ -907,13 +922,13 @@ get_some_switches (void)
 		noreverse = true;
 		break;
 	    case 'o':
-		outfile = savestr (optarg);
+		outfile = xstrdup (optarg);
 		break;
 	    case 'p':
 		strippath = numeric_string (optarg, false, "strip count");
 		break;
 	    case 'r':
-		rejname = savestr (optarg);
+		rejname = xstrdup (optarg);
 		break;
 	    case 'R':
 		reverse = true;
@@ -947,13 +962,13 @@ get_some_switches (void)
 	    case 'Y':
 		if (!*optarg)
 		  fatal ("backup basename prefix is empty");
-		origbase = savestr (optarg);
+		origbase = xstrdup (optarg);
 		break;
 	    case 'z':
 	    case_z:
 		if (!*optarg)
 		  fatal ("backup suffix is empty");
-		origsuff = savestr (optarg);
+		origsuff = xstrdup (optarg);
 		break;
 	    case 'Z':
 		set_utc = true;
@@ -1011,6 +1026,9 @@ get_some_switches (void)
 		else
 		  usage (stderr, 2);
 		break;
+	    case CHAR_MAX + 11:
+		follow_symlinks = true;
+		break;
 	    default:
 		usage (stderr, 2);
 	}
@@ -1019,12 +1037,12 @@ get_some_switches (void)
     /* Process any filename args.  */
     if (optind < Argc)
       {
-	inname = savestr (Argv[optind++]);
+	inname = xstrdup (Argv[optind++]);
 	explicit_inname = true;
 	invc = -1;
 	if (optind < Argc)
 	  {
-	    patchname = savestr (Argv[optind++]);
+	    patchname = xstrdup (Argv[optind++]);
 	    if (optind < Argc)
 	      {
 		fprintf (stderr, "%s: %s: extra operand\n",
@@ -1090,8 +1108,8 @@ locate_hunk (lin fuzz)
     lin min_where = last_frozen_line + 1 - (prefix_context - prefix_fuzz);
     lin max_pos_offset = max_where - first_guess;
     lin max_neg_offset = first_guess - min_where;
-    lin max_offset = (max_pos_offset < max_neg_offset
-			  ? max_neg_offset : max_pos_offset);
+    lin max_offset = MAX(max_pos_offset, max_neg_offset);
+    lin min_offset;
 
     if (!pat_lines)			/* null range matches always */
 	return first_guess;
@@ -1137,7 +1155,10 @@ locate_hunk (lin fuzz)
 	  return 0;
       }
 
-    for (offset = 0;  offset <= max_offset;  offset++) {
+    min_offset = max_pos_offset < 0 ? first_guess - max_where
+	       : max_neg_offset < 0 ? first_guess - min_where
+	       : 0;
+    for (offset = min_offset;  offset <= max_offset;  offset++) {
 	char numbuf0[LINENUM_LENGTH_BOUND + 1];
 	char numbuf1[LINENUM_LENGTH_BOUND + 1];
 	if (offset <= max_pos_offset
@@ -1149,7 +1170,7 @@ locate_hunk (lin fuzz)
 	    in_offset += offset;
 	    return first_guess+offset;
 	}
-	if (0 < offset && offset <= max_neg_offset
+	if (offset <= max_neg_offset
 	    && patch_match (first_guess, -offset, prefix_fuzz, suffix_fuzz)) {
 	    if (debug & 1)
 	      say ("Offset changing from %s to %s\n",
@@ -1220,6 +1241,7 @@ abort_hunk_unified (bool header, bool reverse)
   lin old = 1;
   lin lastline = pch_ptrn_lines ();
   lin new = lastline + 1;
+  char const *c_function = pch_c_function();
 
   if (header)
     {
@@ -1234,7 +1256,7 @@ abort_hunk_unified (bool header, bool reverse)
   print_unidiff_range (rejfp, pch_first () + out_offset, lastline);
   fprintf (rejfp, " +");
   print_unidiff_range (rejfp, pch_newfirst () + out_offset, pch_repl_lines ());
-  fprintf (rejfp, " @@\n");
+  fprintf (rejfp, " @@%s\n", c_function ? c_function : "");
 
   while (pch_char (new) == '=' || pch_char (new) == '\n')
     new++;
@@ -1556,7 +1578,7 @@ init_reject (char const *outname)
   int fd;
   fd = make_tempfile (&TMPREJNAME, 'r', outname, O_WRONLY | binary_transput,
 		      0666);
-  TMPREJNAME_needs_removal = 1;
+  TMPREJNAME_needs_removal = true;
   rejfp = fdopen (fd, binary_transput ? "wb" : "w");
   if (! rejfp)
     pfatal ("Can't open stream for file %s", quotearg (TMPREJNAME));
@@ -1649,6 +1671,33 @@ patch_match (lin base, lin offset, lin prefix_fuzz, lin suffix_fuzz)
     return true;
 }
 
+/* Check if the line endings in the input file and in the patch differ. */
+
+static bool
+check_line_endings (lin where)
+{
+  char const *p;
+  size_t size;
+  bool input_crlf, patch_crlf;
+
+  p = pfetch (1);
+  size = pch_line_len (1);
+  if (! size)
+    return false;
+  patch_crlf = size >= 2 && p[size - 2] == '\r' && p[size - 1] == '\n';
+
+  if (! input_lines)
+    return false;
+  if (where > input_lines)
+    where = input_lines;
+  p = ifetch (where, false, &size);
+  if (! size)
+    return false;
+  input_crlf = size >= 2 && p[size - 2] == '\r' && p[size - 1] == '\n';
+
+  return patch_crlf != input_crlf;
+}
+
 /* Do two lines match with canonicalized white space? */
 
 bool
@@ -1705,7 +1754,7 @@ delete_file_later (const char *name, const struct stat *st, bool backup)
 
   if (! st)
     {
-      if (lstat (name, &st_tmp) != 0)
+      if (stat_file (name, &st_tmp) != 0)
 	pfatal ("Can't get file attributes of %s %s", "file", name);
       st = &st_tmp;
     }
@@ -1757,7 +1806,7 @@ struct file_to_output {
 static gl_list_t files_to_output;
 
 static void
-output_file_later (char const *from, int *from_needs_removal, const struct stat *from_st,
+output_file_later (char const *from, bool *from_needs_removal, const struct stat *from_st,
 		   char const *to, mode_t mode, bool backup)
 {
   struct file_to_output *file_to_output;
@@ -1770,11 +1819,11 @@ output_file_later (char const *from, int *from_needs_removal, const struct stat
   file_to_output->backup = backup;
   gl_list_add_last (files_to_output, file_to_output);
   if (from_needs_removal)
-    *from_needs_removal = 0;
+    *from_needs_removal = false;
 }
 
 static void
-output_file_now (char const *from, int *from_needs_removal,
+output_file_now (char const *from, bool *from_needs_removal,
 		 const struct stat *from_st, char const *to,
 		 mode_t mode, bool backup)
 {
@@ -1791,7 +1840,7 @@ output_file_now (char const *from, int *from_needs_removal,
 }
 
 static void
-output_file (char const *from, int *from_needs_removal,
+output_file (char const *from, bool *from_needs_removal,
 	     const struct stat *from_st, char const *to,
 	     const struct stat *to_st, mode_t mode, bool backup)
 {
@@ -1858,13 +1907,13 @@ output_files (struct stat const *st)
   while (gl_list_iterator_next (&iter, &elt, NULL))
     {
       const struct file_to_output *file_to_output = elt;
-      int from_needs_removal = 1;
+      bool from_needs_removal = true;
       struct stat const *from_st = &file_to_output->from_st;
 
       output_file_now (file_to_output->from, &from_needs_removal,
 		       from_st, file_to_output->to,
 		       file_to_output->mode, file_to_output->backup);
-      if (from_needs_removal)
+      if (file_to_output->to && from_needs_removal)
 	unlink (file_to_output->from);
 
       if (st && st->st_dev == from_st->st_dev && st->st_ino == from_st->st_ino)
@@ -1915,12 +1964,12 @@ fatal_exit (int sig)
 }
 
 static void
-remove_if_needed (char const *name, int *needs_removal)
+remove_if_needed (char const *name, bool *needs_removal)
 {
   if (*needs_removal)
     {
       unlink (name);
-      *needs_removal = 0;
+      *needs_removal = false;
     }
 }
 
